/*
 * mtz2hkl - convert mtz-file to shelx HKL format
 * Copyright (C) 2008 Tim Gruene
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "Converter.h"
#include "LCompletor.h"
#include "myExcepts.h"
#include "defines.h"

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <fstream>
#include <iostream>
#include <stdexcept>
#include <cstdlib>
#include <cstring>

Converter::Converter(const std::string& mtzfile, const std::string& hklfile,
	          		 const std::string& data, const std::string& sigma,
	          		 const std::string& rfree, int value_rfree,
	          		 bool forceNoRfree):
mtzfile_(mtzfile), hklfile_(hklfile), label_data_(data), label_sigma_(sigma),
label_rfree_(rfree), value_rfree_(value_rfree), forceNoRfree_(forceNoRfree)
{
	mtz_ = CMtz::MtzGet(mtzfile_.c_str(), false);
	if(mtz_ == NULL) throw myExcepts::FileIO::MtzOpen(mtzfile.c_str());
	columns_ = new int[NUMCOLS];
	columns_[DATA] = -1;
	columns_[SIGMA] = -1;
	columns_[RFREE] = -1;	
}

Converter::~Converter()
{
	CMtz::MtzFree(mtz_);
	delete[] columns_;
}

/**
 * call all steps required to convert and write data.
 * If @c squareAmplitudes is true, amplitudes are converted to
 * intensities
 * \param squareAmplitudes
 */
int Converter::convert(bool squareAmplitudes)
{
	unsigned int numTestSet;
	try {
		completeLabels();
		readData();

		if(!data_are_intensities_)
		{
			std::cout << "\n"
			<< "*** WARNING *** WARNING *** WARNING *** WARNING *** WARNING ******************\n"
			<< "*  Data are amplitudes as opposed to intensities. It is strongly recommended *\n"
			<< "*  to use shelx with intensities. If you can get hold of your original data, *\n"
			<< "*  please rerun truncate with the option to keep intensities.                *\n"
			<< "*** WARNING *** WARNING *** WARNING *** WARNING *** WARNING ******************\n"
			<< "\n";
		}
		if (squareAmplitudes)
		{
			squareData();
		}
		scaleData();
		numTestSet = writeData();
		printInfo(numTestSet);
	}
	catch(myExcepts::Label::RFree& rfe)
	{
		throw;
	}
	catch(...)
	{
		throw;
	}
	return 0;	
}

int Converter::mtzinfo() const
{
	std::cout << "Columns and types in mtzfile:\n";
	for(int xtal = 0; xtal < mtz_->nxtal; ++xtal)
	{
		for(int set = 0; set < mtz_->xtal[xtal]->nset; ++set)
		{
			for(int col = 0; col < mtz_->xtal[xtal]->set[set]->ncol; ++col)
			{
				std::cout << "  "
					 	<< std::setw(32) 
						<< std::left 
						<< mtz_->xtal[xtal]->set[set]->col[col]->label
						<< std::setw(5) 
						<< mtz_->xtal[xtal]->set[set]->col[col]->type 
						<< '\n';
			}
		}
	}
	return 0;
}

/**
 * ensure all data are within -9999.0 and 9999.0 to fit the fixed format
 * This applies a linear scale to all Amplitudes and their sigmas.
 */
int Converter::scaleData()
{
	double scale (Defs::HKLF::Scale);
	double maxData, minData;
	std::vector<Reflex>::iterator it;
	
	maxData = std::max_element(reflections_.begin(), reflections_.end())->data();
	minData = std::min_element(reflections_.begin(), reflections_.end())->data();
		
	std::abs(minData) > std::abs(maxData)? 
		scale /= std::abs(minData) : scale /= std::abs(maxData);
	for(it = reflections_.begin(); it != reflections_.end(); ++it)
	{
		*it = scale* *it;
	}	
	return 0;
	
}

/**
 * If data are amplitudes, square data to (re-)create intensities.
 * sigmas are converted according to
 * \f$ \sigma_{I} := 2\sigma_{F} F\f$
 */
void Converter::squareData()
{
	std::vector<Reflex>::iterator it;
	if (data_are_intensities_)
	{
		std::cerr << "*** WARNING ***\n"
		<< "   Squaring of data requested (option '-2'), but data are recognised as\n"
		<< "   Intensities. Data will not be squared\n";
		return;
	}
	for (it = reflections_.begin(); it != reflections_.end(); ++it)
	{
		it->square();
	}
	data_are_intensities_ = true;
}
unsigned int Converter::writeData()
{
	std::ofstream output (hklfile_.c_str());
	std::vector<Reflex>::const_iterator it;
	unsigned int totalRfree(0);	// counter for test set reflections
	
	if(! output.is_open())
	{
		std::cout << " An error occurred opening file " << hklfile_ << "\n"
		<< " Please check write permissions and filename.\n";
		throw myExcepts::FileIO::HklOpen(hklfile_);
	}

	 
	for (it = reflections_.begin(); it != reflections_.end(); ++it)
	{
		int rfree;
		if (it->rfree() == value_rfree_)
		{
			rfree = -1;
			++totalRfree;
		}
		else 
		{
			rfree = 1;
		}

		output << std::setw(4) << it->h() 
		       << std::setw(4) << it->k()
		       << std::setw(4) << it->l()
		       << std::fixed
		       << std::setw(8) << std::setprecision(2) << it->data()
		       << std::setw(8) << std::setprecision(2) << it->sigma()
		       << std::setw(4) << rfree << "\n";
	} 
	
	output.close();
	return totalRfree;
}

/**
 * Completion of labels can become quite complex. Therefore this function creates
 * a new class, LCompletor, to take over that task
 */
int Converter::completeLabels()
{
	std::string type_data, type_sigma, type_rfree;
	char* lblData  = NULL;
	char* lblSigma = NULL;
	char* lblRfree = NULL;
	if (! label_data_.empty())	
	{	
		lblData = (char*) std::calloc(sizeof(char), (label_data_.size()+1));
		std::strncpy(lblData, label_data_.c_str(), label_data_.size());
	}
	if (! label_sigma_.empty())	
	{	
		lblSigma = (char*) std::calloc(sizeof(char), label_sigma_.size()+1);
		std::strncpy(lblSigma, label_sigma_.c_str(), label_sigma_.size());
	}
	if (! label_rfree_.empty())	
	{	
		lblRfree = (char*) std::calloc(sizeof(char), label_rfree_.size()+1);
		std::strncpy(lblRfree, label_rfree_.c_str(), label_rfree_.size());
	}
	
	try {
		LCompletor completor(mtz_, forceNoRfree_, lblData, lblSigma, lblRfree);
		columns_[DATA]        = completor.getIndexData() -1;
		columns_[SIGMA]       = completor.getIndexSigma()-1;
		label_data_			  = completor.getLblData();
		label_sigma_		  = completor.getLblSigma();
		data_are_intensities_ = completor.areIntensities();
		columns_[RFREE]       = completor.getIndexRFree()-1;
		label_rfree_		  = completor.getLblRfree();
	}
	catch(myExcepts::Label::RFree&)
	{
		if (!forceNoRfree_)
			{
				free(lblData);
				free(lblSigma);
				free(lblRfree);
				throw;
			}
	}
	catch(myExcepts::Label::Label& lbl)
	{
		free(lblData);
		free(lblSigma);
		free(lblRfree);
		throw;
	}
	catch(...)
	{
		free(lblData);
		free(lblSigma);
		free(lblRfree);
		std::cerr << " ** Unknown Error: Please inform the author \n";
		throw;
	}

	free (lblData);
	free(lblSigma);
	free(lblRfree);
 	return 0;
} 

int Converter::readData()
{
	float in_resl;			         // resolution per reflection
	float inData [mtz_->ncol_read+1];	 // input data from mtz file
	int  inLogmss[mtz_->ncol_read+1]; // flags for validity input
	int h, k, l, rfree;
	float data, sigma;
	int read_reflection(1);			 // ccp4 fortran leftover starts counting at 1

	if(columns_[DATA] == -1 || columns_[SIGMA] == -1 || 
				(columns_[RFREE] < 0 && forceNoRfree_ == false))
		throw std::logic_error("***Converter Error***: Columns are not initialised");
		
	reflections_.clear();
	
	if (columns_[RFREE] < 0)
	{
		while (CMtz::ccp4_lrrefl(mtz_, &in_resl, inData, inLogmss, read_reflection) == 0)
		{
			if (inLogmss[columns_[DATA]]  == 0 && 
			    inLogmss[columns_[SIGMA]] == 0)
			{
				h = int(inData[0]);
				k = int(inData[1]);
				l = int(inData[2]);
				data   = inData[columns_[DATA]];
				sigma  = inData[columns_[SIGMA]];
				rfree  = value_rfree_ + 1;	// just make sure this is different
					
				reflections_.push_back(Reflex(h, k, l, data, sigma, rfree));	
			} 
			
			++read_reflection;
		}

	}
	else
	{
		while (CMtz::ccp4_lrrefl(mtz_, &in_resl, inData, inLogmss, read_reflection) == 0)
		{
			if (inLogmss[columns_[DATA]]  == 0 && 
			    inLogmss[columns_[SIGMA]] == 0 && 
			    inLogmss[columns_[RFREE]] == 0 )
			{
				h = int(inData[0]);
				k = int(inData[1]);
				l = int(inData[2]);
				data   = inData[columns_[DATA]];
				sigma  = inData[columns_[SIGMA]];
				rfree = int(inData[columns_[RFREE]]);
					
				reflections_.push_back(Reflex(h, k, l, data, sigma, rfree));	
			} 
			
			++read_reflection;
		}
	}
	return read_reflection-1;
}

/**
 * Print Information about what has been done
 */
void Converter::printInfo(unsigned int numTestSet) const
{
	std::cout << " ** SUMMARY:\n"
			  << "    Data:  column "
			  << std::setw(3) << columns_[DATA] 
			  << ", label: \"" << label_data_ << "\"\n"
			  << "    Sigma: column "
			  << std::setw(3) << columns_[SIGMA] 
			  << ", label: \"" << label_sigma_ << "\"\n";
	if(columns_[RFREE] >= 0)
	{
		std::cout  << "    Rfree: column "
				   << std::setw(3) << columns_[RFREE] 
				   << ", label: \"" << label_rfree_ << "\"\n";
	}
	std::cout << "    Format:     HKLF " << 3+data_are_intensities_ << "\n"
			  << "    Spacegroup: " << mtz_->mtzsymm.spcgrpname << "\n";
	std::cout << "    Cells present in mtz-file:\n";
	for(int xtal = 0; xtal < mtz_->nxtal; ++xtal)
	{
		std::cout 	<< "        " 
					<< mtz_->xtal[xtal]->cell[0] << ' '
					<< mtz_->xtal[xtal]->cell[1] << ' '
					<< mtz_->xtal[xtal]->cell[2] << "  "
					<< mtz_->xtal[xtal]->cell[3] << ' '
					<< mtz_->xtal[xtal]->cell[4] << ' '
					<< mtz_->xtal[xtal]->cell[5] << "\n";
	}
	
	std::cout << "    " << reflections_.size() 
			  << " reflections written to file \""
	          << hklfile_ << "\"\n";
	  
	if(columns_[RFREE] >= 0)
	{
		std::cout  << "    " << numTestSet << " Reflections with " << label_rfree_ 
				   << " = " << value_rfree_	  << " marked as test set (-1)\n";
	}
	
	if(columns_[RFREE] < 0)
	{
		std::cout << "\n"
				  << " ** WARNING:\n"
		          << "            No RFREE-flag was provided or found, resulting\n"
		          << "            hkl-file is flagged with 1's only\n";
	}
			
	
}
