/*
 * mtz2hkl - convert mtz-file to shelx HKL format
 * Copyright (C) 2008 Tim Gruene
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "LCompletor.h"
#include "utils.h"

#include <stdexcept>
#include <iostream>
#include <cstring>

LCompletor::LCompletor(CMtz::MTZ* mtz, bool ignore_rfree,
									   const char* ldata, 
						 	 	 	   const char* lsigma,
						 		 	   const char* lrfree):
mtz_(mtz), searchLblData_(ldata), searchLblSigma_(lsigma), 
searchLblRfree_(lrfree), data_are_intensities_(true), ignore_rfree_(ignore_rfree)
{	 
	data_.column_index_  = -1;
	sigma_.column_index_ = -1;
	rfree_.column_index_ = -1;
	if (mtz_ == NULL)	
		throw std::runtime_error("*** ERROR *** Empty mtz struct passed to LCompletor");
	
	try {
		searchLabels();
	}
	catch (myExcepts::Label::Label& le)
	{
		throw;
	}
}

LCompletor::~LCompletor()
{
}

void LCompletor::searchLabels() throw (myExcepts::Label::Label)
{
	std::string lbl_cur;
	std::string lbl_sigma;
	
	bool uniqueRfree(false);	// Rfree successfully found
	bool uniqueDataI(false);	// Intensity + sigma successfully found
	bool uniqueDataA(false);	// Amplitude + sigma successfully found
	
	try 
	{
		for(int x = 0; x < mtz_->nxtal; ++x)
		{
			for (int s = 0; s < mtz_->xtal[x]->nset; ++s)
			{
				for(int c = 0; c < mtz_->xtal[x]->set[s]->ncol; ++c)
				{
					CMtz::MTZCOL* this_col = mtz_->xtal[x]->set[s]->col[c];
					CMtz::MTZCOL* next_col = NULL;
					char this_type          = this_col->type[0];
					char next_type			= 0;
					if(c < mtz_->xtal[x]->set[s]->ncol - 1)
					{
						next_col  = mtz_->xtal[x]->set[s]->col[c+1];
						next_type = next_col->type[0];
					}
					
					switch(this_type)
					{
						case 'I':	if(uniqueRfree 	== false)
										uniqueRfree = processI(this_col);
									break;
						case 'J':	if(uniqueDataI 	== false)
										uniqueDataI = processJ(this_col, next_col);
									if(uniqueRfree && uniqueDataI) return;
									break;
						case 'F':	if(uniqueDataI == false && uniqueDataA	== false)
										uniqueDataA = processA(this_col, next_col);
									break;
						default:	break;
					}
				}
			}
		}
	}
	catch(myExcepts::Label::Label& le)
	{
		throw;
	}
	if (!uniqueRfree)
	{
		if (rfree_candidates_.size() == 1)
		{
			rfree_ 		= rfree_candidates_.front();
			uniqueRfree = true;
		}
		if(rfree_candidates_.size() > 1)
		{
			try {
				checkRfree();
			}
			catch (myExcepts::Label::RFree&)
			{
				if (! ignore_rfree_)	throw;
			}
		}
		if(rfree_candidates_.empty() && ! ignore_rfree_)
		{
			throw myExcepts::Label::RFree();
		}
	} 
	
	if(uniqueDataI)		return;
	else if (dataI_candidates_.size() == 1)
	{
		data_  = dataI_candidates_.front();
		sigma_ = sigmaI_candidates_.front();
		return;
	}
	if(uniqueDataA)		return;
	else if (dataF_candidates_.size() == 1)
	{
		data_  = dataF_candidates_.front();
		sigma_ = sigmaF_candidates_.front();
		data_are_intensities_ = false;
		return;
	}
	else if (dataI_wo_Sigma_candidates_.size() == 1)
	{
		data_  = dataI_wo_Sigma_candidates_.front();
		sigmaByName();
		return;
	}
	else if (dataF_wo_Sigma_candidates_.size() == 1)
	{
		data_  = dataF_wo_Sigma_candidates_.front();
		sigmaByName();
		data_are_intensities_ = false;
		return;
	}
	
	throw myExcepts::Label::Label("both Amplitude and Intensity");
}

/**
 * Process an mtz-column entry to for rfree flags
 * \param col	MTZ column type
 */
bool LCompletor::processI(const CMtz::MTZCOL* col)
{
	// if search label is not set, add this column to the candidates
	if(searchLblRfree_ == NULL)
	{
		Column rfree = {col->label, col->type, col->source};
		rfree_candidates_.push_back(rfree);
		return false;
	}	
	// if search label is set and matches, use this column for good 
	else if (strncmp(searchLblRfree_, col->label, 31) == 0)
	{
		rfree_.label_        = col->label;
		rfree_.type_         = col->type;
		rfree_.column_index_ = col->source;
		return true;
	}
	// if search label is set but does not match, continue
	else return false;	
}

/**
 * Check whether this column is or can be a data intensity column
 * \param this_col	actual J-column
 * \param next_col  following column
 */
bool LCompletor::processJ(const CMtz::MTZCOL* this_col, const CMtz::MTZCOL* next_col)
{
	// label is not given -> check if this one has neighbour Q
	if(searchLblData_ == NULL)
	{
		if(next_col == NULL)	return false;
		else if (next_col->type[0] == 'Q')
		{
			Column col1 = {this_col->label, this_col->type, this_col->source};
			dataI_candidates_.push_back(col1);
			Column col2 = {next_col->label, next_col->type, next_col->source};
			sigmaI_candidates_.push_back(col2);
			return false;				
		}
		else 
		{
			Column col1 = {this_col->label, this_col->type, this_col->source};
			dataI_wo_Sigma_candidates_.push_back(col1);
			return false;
		}
	}
	else if (strncmp(searchLblData_, this_col->label, 31) == 0)
	{
		data_.label_    	= this_col->label;
		data_.type_			= this_col->type;
		data_.column_index_	= this_col->source;  
		return processQ(this_col, next_col, true);
	}
	else return false;	
}

/**
 * Check wheterh this column is or can be data amplitude column
 * \param	this_col	actual F column
 * \param	next_col	following column
 */
bool LCompletor::processA(const CMtz::MTZCOL* this_col, const CMtz::MTZCOL* next_col)
{
	// label is not given -> check if this one has neighbour Q
	if(searchLblData_ == NULL)
	{
		if(next_col == NULL)	
			return false;
		else if (next_col->type[0] == 'Q')
		{
			Column col1 = {this_col->label, this_col->type, this_col->source};
			dataF_candidates_.push_back(col1);
			Column col2 = {next_col->label, next_col->type, next_col->source};
			sigmaF_candidates_.push_back(col2);
			return false;				
		}
		else
		{
			Column col1 = {this_col->label, this_col->type, this_col->source};
			dataF_wo_Sigma_candidates_.push_back(col1);
			return false;
		}
	}
	else if(strncmp(searchLblData_, this_col->label, 31) == 0)
	{
		data_.label_		= this_col->label;
		data_.type_			= this_col->type;
		data_.column_index_	= this_col->source;
		return processQ(this_col, next_col, false);
	}
	else return false;
}

/**
 * check whether a sigma label is found for @c this_col.
 * This should only be called from processJ (@c intensity=true)
 * or processA(@c intensity = false), i.e., this_col was already processed
 * \param this_col	col for intensity
 * \param next_col  putative col for sigma
 * \param intensity switch for this_col being intensity(true) or amplitude(false)
 */
bool LCompletor::processQ(const CMtz::MTZCOL* this_col,
						  const CMtz::MTZCOL* next_col,
						  bool intensity )
{
	// 
	if(searchLblSigma_ == NULL)
	{
		if(next_col->type[0] != 'Q')
		{
			return false;
		}
		else	// use this_col and next_col as candidates
		{
			Column col1 = {this_col->label, this_col->type, this_col->source};
			Column col2 = {next_col->label, next_col->type, next_col->source};
			if (intensity == true)
			{
				dataI_candidates_.push_back(col1);
				sigmaI_candidates_.push_back(col2);
				return false;
			}
			else
			{
				dataF_candidates_.push_back(col1);
				sigmaF_candidates_.push_back(col2);
				return false;
			}
		}
	}
	else		// sigma label was provided
	{
		int res = completeColumn(sigma_, searchLblSigma_);
		if(res == true)
		{
			return true;
		}
		else 		// fatal
		{
			throw myExcepts::Label::NotFound(searchLblSigma_);
		}
	}
}

/**
 * scans through all labels in mtz-file. If the correct one is found,
 * the Column @c column is updated by type and column index
 * \param	column	which column to update
 * \param 	label	identifier for column
 * \return	true iff label was found
 */
bool LCompletor::completeColumn(Column& column, const std::string& label)
{
	std::string type;
	unsigned int index;
	std::string  curLbl;
	
	for(int x = 0; x < mtz_->nxtal; ++x)
	{
		for (int s = 0; s < mtz_->xtal[x]->nset; ++s)
		{
			for(int c = 0; c < mtz_->xtal[x]->set[s]->ncol; ++c)
			{
				type     = mtz_->xtal[x]->set[s]->col[c]->type;
				curLbl   = mtz_->xtal[x]->set[s]->col[c]->label;
				index    = mtz_->xtal[x]->set[s]->col[c]->source;
				if (label == curLbl)
				{
					column.label_ = curLbl;
					column.type_  = type;
					column.column_index_ = index;
					return true;
				}
			}
		}
	}
	return false;
}

/**
 * In the case of multiple rfree label candidates, see if there is only one label with a
 * 'free' in its name (i.e., the other labels of type 'I' are something else, 
 * not RFree entries.
 * if this is not the case, throw an exception
 */
void LCompletor::checkRfree() throw(myExcepts::Label::RFree)
{
	Column rfree = {"", "", -1};
	std::vector<Column>::const_iterator it;
	unsigned int num_matches(0);
	
	for(it = rfree_candidates_.begin(); it != rfree_candidates_.end(); ++it)
	{
		if (Utils::String::lowerCase(it->label_).find("free"))
		{
			++num_matches;
			rfree = *it;
		}
	}
	if (num_matches == 1)
	{
		rfree_ = rfree;
		return;
	}
	else	throw myExcepts::Label::RFree();
}

/**
 * In cases where data and sigma are not adjacent, the above processI/ processA miss
 * the sigma entry. As last resort check the labels of 'Q' whether they contain the
 * name of data_ (like in "sigF" or "SIGImine")
 */
void LCompletor::sigmaByName()
{
	std::vector<Column> Qcolumns;
	char type;
	std::string testlabel;
	std::string datalabel (Utils::String::lowerCase(data_.label_));
	
	for(int x = 0; x < mtz_->nxtal; ++x)
	{
		for (int s = 0; s < mtz_->xtal[x]->nset; ++s)
		{
			for(int c = 0; c < mtz_->xtal[x]->set[s]->ncol; ++c)
			{
				CMtz::MTZCOL* this_col = mtz_->xtal[x]->set[s]->col[c];
				
				type      = this_col->type[0];
				if (type != 'Q') 	continue;
				testlabel = this_col->label;
				testlabel = Utils::String::lowerCase(testlabel);
				if(testlabel.find(datalabel) != std::string::npos)
				{
					Column col = {this_col->label, this_col->type, this_col->source};
					Qcolumns.push_back(col);
				}
			}
		}
	}
	if (Qcolumns.size() == 1)
	{
		sigma_ = Qcolumns.front();
		return;
	}	
	else
		throw myExcepts::Label::RFree();

}
