#ifndef LCOMPLETOR_H_
#define LCOMPLETOR_H_

/*
 * mtz2hkl - convert mtz-file to shelx HKL format
 * Copyright (C) 2008 Tim Gruene
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "myExcepts.h"

#include <ccp4/cmtzlib.h>
#include <string>
#include <vector>

/**
 * \class LCompletor
 * \brief complete or find labels from an mtz-struct
 * 
 * Intelligently finding labels in an mtz-file can become quite 
 * ticky given that column types are not unique and column labels
 * can be chagned by users at will.
 */
class LCompletor
{
private:
	//! summarising struct to store characteristics of a column
	struct Column {
		//! user defined name
		std::string label_;
		//! type of column
		std::string type_;
		//! index of column within mtz-data
		int  column_index_;
	} column_;
	
	CMtz::MTZ* mtz_;
	
	Column data_;	///< column for data (type J or F)
	Column sigma_;	///< column for sigma of data (type Q)
	Column rfree_;	///< column for RFree (type I)
	
	const char* searchLblData_;
	const char* searchLblSigma_;
	const char* searchLblRfree_;
	
	bool  data_are_intensities_;
	bool  ignore_rfree_;
	
	std::vector<Column> dataI_candidates_;
	std::vector<Column> sigmaI_candidates_;
	std::vector<Column> dataF_candidates_;
	std::vector<Column> sigmaF_candidates_;
	std::vector<Column> dataI_wo_Sigma_candidates_;  // data entries without adjacent Sigmas
	std::vector<Column> dataF_wo_Sigma_candidates_;  // data entries without adjacent Sigmas
	std::vector<Column> rfree_candidates_;	
	
	//! check all labels in mtz-struct
	void searchLabels() throw (myExcepts::Label::Label);
	
	//! process a column type I
	bool processI(const CMtz::MTZCOL* );
	
	bool processJ(const CMtz::MTZCOL*, const CMtz::MTZCOL*);
	
	bool processA(const CMtz::MTZCOL*, const CMtz::MTZCOL*);
	
	bool processQ(const CMtz::MTZCOL*, const CMtz::MTZCOL*, bool intensity = true);
	
	//! complete Column entry by label 
	bool completeColumn(Column& column, const std::string& label);
	
	//! search for Rfree in list of Column's
	void checkRfree() throw(myExcepts::Label::RFree);
	
	//! search for sigma labels containing the data_ string
	void sigmaByName();
	
public:
	LCompletor(CMtz::MTZ* mtz, bool ignore_rfree,
							   const char* ldata  = NULL, 
						 	   const char* lsigma = NULL,
						 	   const char* lrfree = NULL
						 	   );
	virtual ~LCompletor();
		
	int getIndexData()  const { return data_.column_index_; }
	int getIndexSigma() const { return sigma_.column_index_; }
	int getIndexRFree() const { return rfree_.column_index_; }

	std::string getLblData()  const { return data_.label_; }
	std::string getLblSigma() const { return sigma_.label_; }
	std::string getLblRfree() const { return rfree_.label_; }

	std::string getTypeData()  const { return data_.type_; }
	std::string getTypeSigma() const { return sigma_.type_; }
	std::string getTypeRfree() const { return rfree_.type_; }

	bool areIntensities() const { return data_are_intensities_; }
};

#endif /*LCOMPLETOR_H_*/
