/*
 * mtz2hkl - convert mtz-file to shelx HKL format
 * Copyright (C) 2008 Tim Gruene
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**
 * \file 	main.cpp
 * \date 	2/4/2007
 * \author 	Tim Gruene
 */
#include "Reflex.h"
#include "Converter.h"
#include "myExcepts.h"
#include "defines.h"


#include <iostream>
#include <string>
#include <sstream>
#include <stdexcept>

#include <getopt.h>

/**
 * long help for user
 */
void usage();
int  str2int(const std::string&);

//! replace .hkl suffix with .mtz
std::string replace_mtz_with_hkl(std::string& mtzname);

int main(int argc, char* argv[])
{
	int result(-1);
	int opt;
	int value_rfree(0);
	bool forceNoRfree (false);    //! force writing in absence of RFree flag
	bool squareAmplitudes(false);
	std::string label_data, label_sigma, label_rfree;
	std::string mtzin, userout, hklout;
	
	/**
	 * Possible option:
	 * h: print help and exit 
	 * f: force conversion even in the absence of an Rfree flag
	 * i: Input mtz-file
	 * d: Label for data (string) [Optional]
	 * s: Label for sigma (string)
	 * r: Label for Rfree-flag (string)
	 * R: Value for test set (number)
	 * o: Filename for output [Optional]
	 * 2: square data in case it is amplitudes (force HKLF4 output)
	 */
	opterr = 0;
	while ( (opt = getopt(argc, argv, "?hd:fo:s:r:R:2")) != -1)
	{
		switch(opt)
		{
			case 'h': case '?':
						usage();
						return EXIT_SUCCESS;
			case 'f':   forceNoRfree = true;
						break;
			case 'd':	label_data 		= optarg;
						break;
			case 'o':	userout			= optarg;
						break;
			case 's':   label_sigma		= optarg;
						break;
			case 'r':	label_rfree		= optarg;
						break;
			case 'R':	try {
							value_rfree		= str2int(optarg);
							break;
						}
						catch(std::runtime_error& e)
						{
							std::cout << "*** Error ***\n"
							 << " Option '-R' requires a number as argument\n"
							 << " " << optarg << " could not be converted\n";
							 usage();
							 return EXIT_FAILURE;
						}
			case '2':	squareAmplitudes = true;
						break;
			default:	usage();
						return EXIT_FAILURE;
		}
	}
	if(optind == argc)
	{
		usage();
		return EXIT_FAILURE;
	}
	
	// convert each file in filelist
	for(int i = optind; i < argc; ++i)
	{
		mtzin  = argv[i];
		// if the users gives '-o' with several files, this will overwrite all 
		// previous conversions
		if(userout.empty())
			hklout = replace_mtz_with_hkl(mtzin);
		else 
			hklout = userout;
		Converter conv(mtzin, hklout, 
					   label_data, label_sigma,
					   label_rfree, value_rfree, forceNoRfree);
		try {
			result = conv.convert(squareAmplitudes);
		}
		catch(myExcepts::Label::RFree& rfe)
		{
			std::cout << " *** ERROR: No suitable RFREE flag found. If you still want to\n"
			          << "            convert this file, please re-run this program with\n"
			          << "            with the '-f' option or provide the RFREE flag with\n"
			          << "            the '-r rfree_flag' option.\n\n";
			conv.mtzinfo();
			return EXIT_FAILURE;
		}
		catch(...)
		{
	
			if(result != 0)
			{
				std::cerr << "*** Error: Conversion to mtz failed.\n"
					<< "    Please explicitly select your labels with the options\n"
					<< "    -d (data), -s (their sigma), -r (rfree flag)\n"
					<< "    Available labels are:\n";
				conv.mtzinfo();
			}
		}
	}
	return result;
}

/**
 * Write out concise information about how to use this program
 */
void usage()
{
	std::cout << " This is " << Defs::Program::name
			  << ", version " << Defs::Program::version
			  << " by " << Defs::Author::name 
			  << ", Copyright " << Defs::Program::copyright << "\n";
 	std::cout << " * Usage:\n"
 	             "   mtz2hkl [options] mtz_input\n\n"
 	             "   where [options] include\n"
 	             "       -d/ -s string labels for data and their sigma\n"
 	             "       -r string     label for RFree\n"
 	             "       -R number     value of  RFree (default: 0)\n"
 	             "       -o string     output written to string instead of mtz_input.hkl\n"
 	             "       -f            force conversion in absence of RFREE\n"
				 "       -2            convert amplitudes to intensities (if applicable)\n"
 	             "\n"
 				 " * Output file name is the same as input but with .mtz replaced by .hkl\n"
 	             " * Output format is HKLF4 with intensities and HKLF3 with amplitude\n"
 	             " * try 'mtz2hkl input.mtz'"
 	             " if you are working with a standard mtz-file\n"
 	             "   to let the program guess the correct labels\n\n";
}

/**
 * replace trailing .mtz with .hkl
 * If .mtz was not found at end of file, it is appended @c mtzname
 */
std::string replace_mtz_with_hkl(std::string& mtzname)
{
	unsigned int period = mtzname.rfind(".mtz");
	
	if(period == std::string::npos)
		mtzname += ".mtz";
	period = mtzname.rfind(".mtz");
	std::string hklname(mtzname);

	return hklname.replace(period, 4, ".hkl");
}

/** convert string to integer **/
int str2int(const std::string& str)
{
	int result;
	std::istringstream stream(str);
	stream >> result;
	if(stream.fail())	throw std::runtime_error("Conversion error: " + str);
	else return result;
}
